/**
 * $Id: game.cc,v 1.18 2003/04/09 03:09:25 i Exp $
 *
 * Copyright (C) shinichiro.h <g940455@mail.ecc.u-tokyo.ac.jp>
 *  http://user.ecc.u-tokyo.ac.jp/~g940455/wp/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "game.h"
#include "player.h"
#include "enemy.h"
#include "shot.h"
#include "bpo.h"
#include "resource.h"
#include "cpputil.h"
#include "mymath.h"
#include "orchestra.h"
#include "padinput.h"
#include "enemy_command.h"
#include "bulletml/bulletmlrunner.h"
#include "effect.h"
#include "profile.h"
#include "clear_screen.h"
#include "bullet.h"
#include "score.h"
#include "message.h"
#include "bonus.h"
#include "music.h"
#include "cpu.h"
#include "cpuinfo_bpo.h"

#include "mysdl.h"
#include <SDL_gfxPrimitives.h>

#include <algorithm>
#include <fstream>
#include <vector>
#include <queue>
#include <list>

struct GameImpl {
	explicit GameImpl(Game* game);
	void run();
	~GameImpl();

	void initCharactors();

	void runPlayer();
	void runEnemy();
	void runBullet();
	void runShot();
	void runEffect();
	void runMessage();
	void processBonus();

	template <class T_>
	void collisionEvent(T_& shots);

	void processReserved();

	void drawPlayer();
	void drawEnemy();
	void drawBullet();
	void drawShot();
	void drawEffect();
	void drawMessage();
	void drawScore();

	void clearScreen();

	void reserveShot(class BulletMLParser* bulletml, int turn);

	void addBullet(class BulletMLState* state,
				   Num x, Num y, double dir, double spd);
	void addShot(Num x, Num y, double dir, double spd);

	bool shouldDestroyBullet(Bullet* bullet) const;
	bool shouldDestroyShot() const;

	void addBonus(Bonus& bonus);

	SDL_Surface* createSpriteSurface(Uint32 size);

	std::auto_ptr<Resource> getResult() const;

	Game* game_;

	Player* player_;
	std::vector<Enemy*> enemys_;
	std::list<Bullet*> bullets_;
	std::deque<Shot*> shots_;
	std::vector<Effect*> bangEffects_;
	std::vector<Effect*> buzzEffects_;
	std::deque<Message*> messages_;

	int turn_;
	Uint32 startTicks_;
	bool end_;

	SDL_Surface* surface_;

	SDL_Surface* enemySurf_;
	SDL_Surface* shotSurf_;
	SDL_Surface* fireBulletSurf_;
	SDL_Surface* bulletSurf_;
	SDL_Surface* bangEffectSurf_;
	SDL_Surface* buzzEffectSurf_;

	std::auto_ptr<Orchestra> orchestra_;

	typedef std::pair<class BulletMLParser*, int> ReservedShot;
    std::queue<ReservedShot> reservedShots_;
    SDL_mutex* reserveMutex_;
	size_t nextShot_;

	int bulletNum_;

	Num buzzLength_, buzzLength2_;
	Num collLength_, collLength2_;
	Num riskyLength2_;
	int buzzEffectNum_;
	int collEffectNum_;

	int clearMode_;
	bool isCpuMode_;

	std::auto_ptr<Score> score_;

	int rank_;
	int rankMax_;
	int missRankDown_;
	int prevRankChangeTurn_;

	int prevDamagedTurn_;

	int prevNotHighSpeedTurn_;
	int prevNotLowSpeedTurn_;

	int prevBuzzTurn_;

	bool riskyFlag_;

	Axis prevAxis_;
	bool prevButton_;
	int prevActionTurn_;
	int prevFireTurn_;

	std::auto_ptr<BonusMgr> bonusMgr_;

	std::auto_ptr<CpuInputBase> cpu_;

	const Music* music_;

};

GameImpl::GameImpl(Game* game)
	: game_(game), turn_(0), end_(false),
	  surface_(BPO::getSurface()), nextShot_(0), bulletNum_(0),
	  prevRankChangeTurn_(0), prevDamagedTurn_(0),
	  prevNotHighSpeedTurn_(0), prevNotLowSpeedTurn_(0),
	  prevBuzzTurn_(0), riskyFlag_(false),
	  prevActionTurn_(0), prevFireTurn_(0),
	  bonusMgr_(new BonusMgr())
{
	initCharactors();

	reserveMutex_ = SDL_CreateMutex();

	music_ = BPO::getMusic();

	orchestra_.reset(new Orchestra(music_->filename(), game));

	buzzLength_ = BPO::rc().get<Num>("buzzLength");
	buzzLength2_ = buzzLength_ * buzzLength_;
	collLength_ = BPO::rc().get<Num>("collisionLength");
	collLength2_ = collLength_ * collLength_;
	riskyLength2_ = collLength2_ * BPO::rc().get<double>("riskyRatio");
	buzzEffectNum_ = BPO::rc().get<int>("buzzEffectNum");
	collEffectNum_ = BPO::rc().get<int>("collisionEffectNum");
	std::string difficulty = BPO::rc().get<std::string>("difficulty");
	rank_ = BPO::rc().get<int>("gameRank" + difficulty);
	rankMax_ = BPO::rc().get<int>("gameRankMax" + difficulty);
	missRankDown_ = BPO::rc().get<int>("gameMissRankDown" + difficulty);

	enemySurf_ = createSpriteSurface(21);
	shotSurf_ = createSpriteSurface(7);
	fireBulletSurf_ = createSpriteSurface(7);
	bulletSurf_ = createSpriteSurface(7);
	bangEffectSurf_ = createSpriteSurface(3);
	buzzEffectSurf_ = createSpriteSurface(3);

	filledCircleRGBA(enemySurf_, 10, 10, 10, 255,255,255,255);
	filledCircleRGBA(shotSurf_, 3, 3, 4, 255,255,255,255);
	filledCircleRGBA(bulletSurf_, 3, 3, 4, 200,200,255,255);
	filledCircleRGBA(fireBulletSurf_, 3, 3, 4, 0,255,0,255);
	filledCircleRGBA(bangEffectSurf_, 1, 1, 2, 255,0,0,255);
	filledCircleRGBA(buzzEffectSurf_, 1, 1, 2, 255,255,0,255);

	clearMode_ = BPO::rc().get<int>("clearMode");

	score_.reset(new Score());
	score_->setRank(rank_);

	BPO::timerStart();

	startTicks_ = SDL_GetTicks();

	PROFILE_INIT;
}

SDL_Surface* GameImpl::createSpriteSurface(Uint32 size) {
	SDL_Surface* swap = 
		MySDL_CreateRGBSurface(surface_->flags, size, size, 8);
	SDL_Surface* ret = SDL_DisplayFormat(swap);
	SDL_FreeSurface(swap);

	SDL_SetColorKey(ret, SDL_SRCCOLORKEY|SDL_RLEACCEL,
					SDL_MapRGB(ret->format, 0, 0, 0));

	return ret;
}

void GameImpl::initCharactors() {
	Resource& rc = BPO::rc();

	isCpuMode_ = rc.get<bool>("cpu");
	if (isCpuMode_) {
		CpuInputBase::setCpuInfomation(
			std::auto_ptr<CpuInfo>(new CpuInfoBpo(game_)));
		cpu_.reset(CpuInputBase::getDefaultCpu().release());
		player_ = new Player(cpu_.get());
	}
	else {
		player_ = new Player(BPO::getInput());
	}
	player_->setX(rc.get<Uint32>("playerX"));
	player_->setY(rc.get<Uint32>("playerY"));

	size_t enemyNum = rc.get<size_t>("enemyNum");
	Uint32 enemyCenterX = rc.get<Uint32>("enemyCenterX");
	Uint32 enemyCenterY = rc.get<Uint32>("enemyCenterY");
	Uint32 enemyEllipseX = rc.get<Uint32>("enemyEllipseX");
	Uint32 enemyEllipseY = rc.get<Uint32>("enemyEllipseY");
	for (size_t i = 0; i < enemyNum; i++) {
		Enemy* e = new Enemy();
		e->setSpeed(i+10);
		e->setX(enemyCenterX +
				std::cos(PI * 2 * i / enemyNum) * enemyEllipseX);
		e->setY(enemyCenterY +
				std::sin(PI * 2 * i / enemyNum) * enemyEllipseY);
		enemys_.push_back(e);
	}
	Enemy::setBaseSpeed(rc.get<Num>("enemyBaseSpeed"));
	Enemy::setCenterX(enemyCenterX);
	Enemy::setCenterY(enemyCenterY);
	Enemy::setEllipseX(enemyEllipseX);
	Enemy::setEllipseY(enemyEllipseY);
}

void GameImpl::run() {
	turn_++;

	{
		PROFILE("move");
		runPlayer();
		runEnemy();
		runBullet();
		runShot();
		runEffect();
		runMessage();
		processBonus();
	}

	collisionEvent(bullets_);
	collisionEvent(shots_);

	processReserved();

	if (!BPO::shouldSkip()) {
		PROFILE("draw");
		drawPlayer();
		drawEnemy();
		drawBullet();
		drawShot();
		drawEffect();
		drawMessage();
		drawScore();
		//clearScreen();
	}

	end_ = orchestra_->isEnd();
}

void GameImpl::addBonus(Bonus& bonus) {
	bonus.addScore(score_.get());
	messages_.push_back(new Message(BPO::getBMPFont(), bonus.label(),
									player_->x(), player_->y(),
									rnd(2.0)-1, rnd(2.0)-1, bonus.flash()));
}

void GameImpl::processBonus() {
	PROFILE("process bonus");

	if (turn_ - prevRankChangeTurn_ > 60) {
		if (rank_ < rankMax_) rank_++;
		score_->setRank(rank_);
		prevRankChangeTurn_ = turn_;
	}

	int diff = turn_ - prevDamagedTurn_;
	if (diff != 0) {
		if (diff % 3600 == 0) {
			addBonus(bonusMgr_->minuteSurvival());
		}
		else if (diff % 600 == 0) {
			addBonus(bonusMgr_->tenSecondsSurvival());
		}
	}

	if (player_->button()) prevNotHighSpeedTurn_ = turn_;
	else prevNotLowSpeedTurn_ = turn_;

	if (turn_ - prevNotHighSpeedTurn_ > 60 * 30) {
		addBonus(bonusMgr_->highSpeedDodge());
		prevNotHighSpeedTurn_ = turn_;
	}
	else if (turn_ - prevNotLowSpeedTurn_ > 60 * 30) {
		addBonus(bonusMgr_->lowSpeedDodge());
		prevNotLowSpeedTurn_ = turn_;
	}

	if (turn_ - prevBuzzTurn_ > 60 * 10) {
		addBonus(bonusMgr_->safetyDodge());
		prevBuzzTurn_ = turn_;
	}

	if (riskyFlag_ && turn_ - prevBuzzTurn_ > 2) {
		addBonus(bonusMgr_->riskyDodge());
		riskyFlag_ = false;
	}

	bool good = (turn_ - prevFireTurn_) < 4;
	bool just = (turn_ - prevFireTurn_) < 1;
    SDL_mutexP(reserveMutex_);
    {
		if (!reservedShots_.empty()) {
			int diff = reservedShots_.front().second - turn_;
			good |= diff < 4;
			just |= diff < 1;
		}
	}
    SDL_mutexV(reserveMutex_);

	if (player_->axis() != prevAxis_)
	{
		if (turn_ - prevActionTurn_ > 12) {
			if (just) {
				addBonus(bonusMgr_->justTiming());
			}
			else if (good) {
				addBonus(bonusMgr_->goodTiming());
			}
		}
		prevActionTurn_ = turn_;
	}

	prevButton_ = player_->button();
	prevAxis_ = player_->axis();
}

void GameImpl::drawScore() {
	PROFILE("draw score");
	score_->draw();
}

void GameImpl::clearScreen() {
	{
		PROFILE("flip");
		SDL_UpdateRect(surface_, 0,0,0,0);
	}

	PROFILE("clear");

	if (SDL_MUSTLOCK(surface_)) SDL_LockSurface(surface_);

	::clearScreen(surface_, clearMode_);

	if (SDL_MUSTLOCK(surface_)) SDL_UnlockSurface(surface_);
}

void GameImpl::runPlayer() {
	PROFILE("move player");

	player_->move();

	if (player_->x() < 4) player_->setX(4);
	else if (player_->x() > surface_->w-4) player_->setX(surface_->w-4);
	if (player_->y() < 4) player_->setY(4);
	else if (player_->y() > surface_->h-4) player_->setY(surface_->h-4);
}

void GameImpl::runEnemy() {
	PROFILE("move enemy");

	for (size_t i = 0; i < enemys_.size(); i++) {
		Num r = enemys_[i]->speed() * PI * turn_ * Enemy::baseSpeed();
		enemys_[i]->setX(Enemy::centerX() + std::cos(r) * Enemy::ellipseX());
		enemys_[i]->setY(Enemy::centerY() + std::sin(r) * Enemy::ellipseY());
	}
}

void GameImpl::runBullet() {
	PROFILE("move bullet");

	std::for_each(bullets_.begin(), bullets_.end(),
				  std::mem_fun(&Bullet::event));

	for (std::list<Bullet*>::iterator ite = bullets_.begin();
		 ite != bullets_.end();)
	{
		if (shouldDestroyBullet(*ite)) {
			delete *ite;
			ite = bullets_.erase(ite);
		}
		else {
			++ite;
		}
	}
}

void GameImpl::runShot() {
	PROFILE("move shot");

	std::for_each(shots_.begin(), shots_.end(), std::mem_fun(&Shot::move));

	while (shouldDestroyShot()) {
		delete shots_.front();
		shots_.pop_front();
	}
}

void GameImpl::runEffect() {
	PROFILE("move effect");

	if (!buzzEffects_.empty()) {
		delete buzzEffects_.back();
		buzzEffects_.pop_back();

		std::for_each(buzzEffects_.begin(), buzzEffects_.end(),
					  std::mem_fun(&Effect::move));
	}
	if (!bangEffects_.empty()) {
		delete bangEffects_.back();
		bangEffects_.pop_back();

		std::for_each(bangEffects_.begin(), bangEffects_.end(),
					  std::mem_fun(&Effect::move));
	}
}

void GameImpl::runMessage() {
	PROFILE("move message");

	if ((!messages_.empty()) && messages_.front()->getTurn() > 100) {
		delete messages_.front();
		messages_.pop_front();
	}

	std::for_each(messages_.begin(), messages_.end(),
				  std::mem_fun(&Message::move));
}

template <class T_>
void GameImpl::collisionEvent(T_& shots) {
	PROFILE("collision event");

	if (player_->isInvisible()) return;
	Num x = player_->x();
	Num y = player_->y();

	for (typename T_::const_iterator ite = shots.begin();
		 ite != shots.end(); ++ite)
	{
		if (std::abs(x - (*ite)->x()) < buzzLength_ &&
			std::abs(y - (*ite)->y()) < buzzLength_)
		{
			Num dx = (*ite)->x() - x;
			Num dy = (*ite)->y() - y;
			Num d = length(dx, dy);
			if (d < collLength2_) {
				player_->setInvisible();

				delete_clear(bangEffects_);
				bangEffects_.resize(collEffectNum_);
				for (int i = 0; i < collEffectNum_; i++) {
					bangEffects_[i] = new Effect(x, y, rnd(6.0)-3, rnd(6.0)-3);
				}

				std::string msg(
					boost::lexical_cast<std::string>(score_->buzz()) +
					" Buzz");
				messages_.push_back(new Message(
										BPO::getBMPFont(), msg,
										x, y, rnd(2.0)-1, rnd(2.0)-1,
										score_->buzz() >= 100));

				score_->doDamage();

				rank_ -= missRankDown_;
				if (rank_ < 0) rank_ = 0;
				prevRankChangeTurn_ = turn_;
				prevDamagedTurn_ = turn_;
				prevNotHighSpeedTurn_ = turn_;
				prevNotLowSpeedTurn_ = turn_;
				riskyFlag_ = false;

				break;
			}
			else if (d < buzzLength2_) {
				if (!(*ite)->isBuzzed()) {
					(*ite)->setBuzzed();

					for (int i = 0; i < buzzEffectNum_; i++) {
						buzzEffects_.push_back(
							new Effect((*ite)->x(), (*ite)->y(),
									   dx * rnd(1.0), dy * rnd(1.0)));
					}

					score_->doBuzz();

					if (score_->buzz() % 10 == 0) {
						std::string msg(
							boost::lexical_cast<std::string>(score_->buzz()) +
							" Buzz");
						messages_.push_back(new Message(
												BPO::getBMPFont(), msg,
												x, y, rnd(2.0)-1, rnd(2.0)-1,
												score_->buzz() % 100 == 0));
					}
				}
				prevBuzzTurn_ = turn_;

				if (d < riskyLength2_) riskyFlag_ = true;
			}
		}
	}
}

bool GameImpl::shouldDestroyBullet(Bullet* bullet) const {
	if (bullet->isDead()) return true;

	int x = bullet->xi();
	int y = bullet->yi();
	return (x < 0 || y < 0 || x > surface_->w || y > surface_->h);
}

bool GameImpl::shouldDestroyShot() const {
	if (shots_.empty()) return false;

	int x = shots_[0]->xi();
	int y = shots_[0]->yi();
	return (x < 0 || y < 0 || x > surface_->w || y > surface_->h);
}

void GameImpl::processReserved() {
	PROFILE("process reserved");

	BulletMLParser* bullet = 0;

    SDL_mutexP(reserveMutex_);
    {
        while (!reservedShots_.empty() &&
               reservedShots_.front().second <= turn_)
        {
			bullet = reservedShots_.front().first;
			reservedShots_.pop();
        }
    }
    SDL_mutexV(reserveMutex_);

	if (bullet) {
		EnemyCommand com(game_, bullet, enemys_[nextShot_]);
		com.run();
		nextShot_ = (nextShot_ + 1) % 7;

		prevFireTurn_ = turn_;
	}
}

void GameImpl::drawPlayer() {
	PROFILE("draw player");

	int x = player_->xi();
	int y = player_->yi();

	circleRGBA(surface_, x+rnd(5)-2, y+rnd(5)-2, 15+rnd(5), 0, 127, 0, 255);
	circleRGBA(surface_, x+rnd(5)-2, y+rnd(5)-2, 15+rnd(5), 0, 127, 0, 255);
	circleRGBA(surface_, x+rnd(5)-2, y+rnd(5)-2, 15+rnd(5), 0, 127, 0, 255);

	filledCircleRGBA(surface_, x, y, 4, 255, 0, 0, 255);
}

void GameImpl::drawEnemy() {
	PROFILE("draw enemy");

	Uint8 c = turn_ & 255;
	if (c > 127) c = 255 - c;
	c += 127;
	for (size_t i = 0; i < enemys_.size(); i++) {
		MySDL_BlitSurface(enemySurf_, surface_,
						  enemys_[i]->xi() - 10, enemys_[i]->yi() - 10);
	}
}

void GameImpl::drawBullet() {
	PROFILE("draw bullet");

	for (std::list<Bullet*>::iterator ite = bullets_.begin();
		 ite != bullets_.end(); ++ite)
	{
		SDL_Surface* srcSurf;
		if ((*ite)->canFire()) {
			srcSurf = fireBulletSurf_;
		}
		else {
			srcSurf = bulletSurf_;
		}
		MySDL_BlitSurface(srcSurf, surface_,
						  (*ite)->xi() - 3, (*ite)->yi() - 3);
	}
}

void GameImpl::drawShot() {
	PROFILE("draw shot");

	for (size_t i = 0; i < shots_.size(); i++) {
		MySDL_BlitSurface(shotSurf_, surface_,
						  shots_[i]->xi() - 3, shots_[i]->yi() - 3);
	}
}

void GameImpl::drawEffect() {
	PROFILE("draw effect");

	for (size_t i = 0; i < bangEffects_.size(); i++) {
		MySDL_BlitSurface(bangEffectSurf_, surface_,
						  bangEffects_[i]->xi()-1, bangEffects_[i]->yi()-1);
	}
	for (size_t i = 0; i < buzzEffects_.size(); i++) {
		MySDL_BlitSurface(buzzEffectSurf_, surface_,
						  buzzEffects_[i]->xi()-1, buzzEffects_[i]->yi()-1);
	}
}

void GameImpl::drawMessage() {
	PROFILE("draw message");

	std::for_each(messages_.begin(), messages_.end(),
				  std::bind2nd(std::mem_fun(&Message::render), surface_));
}

void GameImpl::reserveShot(class BulletMLParser* bulletml, int turn) {
    SDL_mutexP(reserveMutex_);
    {
        reservedShots_.push(ReservedShot(bulletml, turn));
    }
    SDL_mutexV(reserveMutex_);
}

void GameImpl::addShot(Num x, Num y, double dir, double spd) {
	bulletNum_++;
	Num vx = spd * std::sin(dir);
	Num vy = - spd * std::cos(dir);
	shots_.push_back(new Shot(x, y, vx, vy));

	if (isCpuMode_)
		cpu_->registShot(std::make_pair(Point(x, y), Point(vx, vy)));
}

void GameImpl::addBullet(class BulletMLState* state,
						 Num x, Num y, double dir, double spd)
{
	bulletNum_++;
	Num vx = spd * std::sin(dir);
	Num vy = - spd * std::cos(dir);
	bullets_.push_back(new Bullet(game_, state, x, y, vx, vy));

	if (isCpuMode_)
		cpu_->registShot(std::make_pair(Point(x, y), Point(vx, vy)));
}

GameImpl::~GameImpl() {
	PROFILE_OUTPUT;

	// stop audio first.
	orchestra_.reset(0);

	SDL_FreeSurface(enemySurf_);
	SDL_FreeSurface(shotSurf_);
	SDL_FreeSurface(bulletSurf_);
	SDL_FreeSurface(fireBulletSurf_);
	SDL_FreeSurface(bangEffectSurf_);
	SDL_FreeSurface(buzzEffectSurf_);

	delete player_;
	delete_clear(enemys_);
	delete_clear(bullets_);
	delete_clear(shots_);
	delete_clear(bangEffects_);
	delete_clear(buzzEffects_);
}

Game::Game()
	: pimpl_(new GameImpl(this))
{
	pimpl_->orchestra_->start();
}

void Game::run() {
	pimpl_->run();
}

void Game::addShot(Num x, Num y, double dir, double spd) {
	pimpl_->addShot(x, y, dir, spd);
}

void Game::addBullet(BulletMLState* state,
					 Num x, Num y, double dir, double spd)
{
	pimpl_->addBullet(state, x, y, dir, spd);
}

void Game::reserveShot(BulletMLParser* bulletml, int turn) {
	pimpl_->reserveShot(bulletml, turn);
}

int Game::getTurn() const {
	return pimpl_->turn_;
}

int Game::getRank() const {
	return pimpl_->rank_;
}

Player* Game::getPlayer() {
	return pimpl_->player_;
}

std::vector<Charactor::HaveSpeed*> Game::getBullets() {
	std::vector<Charactor::HaveSpeed*> ret;
	std::copy(pimpl_->bullets_.begin(), pimpl_->bullets_.end(),
			  std::back_inserter(ret));
	std::copy(pimpl_->shots_.begin(), pimpl_->shots_.end(),
			  std::back_inserter(ret));
	return ret;
}

bool Game::isEnd() const {
	return pimpl_->end_;
}

Game::~Game() {
	delete pimpl_;
}

/// $B%O%$%9%3%"=hM}$b$d$C$F$$$k!#:G0-!#(B
std::auto_ptr<Resource> GameImpl::getResult() const {
	std::auto_ptr<Resource> ret(new Resource());

	double second = (double)(SDL_GetTicks() - startTicks_) / 1000;

	ret->set("score", score_->score());
	ret->set("maxBuzz", score_->maxBuzz());
	ret->set("buzzScore", score_->buzzScore());
	ret->set("buzzCount", score_->buzzNum());
	ret->set("buzzPercent",
			 (bulletNum_) ? score_->buzzNum() * 100 / bulletNum_ : 0);
	ret->set("second", (int)second);
	ret->set("fps", BPO::getFps());
	int bulletPerSecond = (int)(bulletNum_ / second);
	ret->set("bulletPerSecond", bulletPerSecond);

	int finalScore = (end_ && second > 60)
		? (int)(score_->score() * 60 / second) : 0;
	ret->set("finalScore", finalScore);

	std::ifstream is("save/highscore.dat");
	int highScore = 0;
	if (is.is_open()) {
		is >> highScore;
	}
	is.close();
	if (finalScore > highScore && !isCpuMode_ && bulletPerSecond > 1) {
		ret->set("highscored", "High Scored");
		std::ofstream os("save/highscore.dat");
		os << finalScore << std::endl
		   << *music_
		   << *bonusMgr_
		   << *ret;
	}
	else {
		ret->set("highscored", "");
	}

	return ret;
}

std::auto_ptr<Resource> Game::getResult() const {
	return pimpl_->getResult();
}

std::auto_ptr<BonusMgr> Game::getBonus() const {
	return pimpl_->bonusMgr_;
}

std::auto_ptr<Music> Game::getMusic() const {
	return std::auto_ptr<Music>(new Music(*pimpl_->music_));
}
