#include "orchestra.h"
#include "bpo.h"
#include "cpputil.h"
#include "game.h"
#include "resource.h"

#include <math.h>
#include <fstream>
#include <set>

extern void init_genrand(unsigned long s);

#ifdef BORLAND
extern "C" {
	extern int Mix_SetMusicCMDPipe(const char *command, Uint16 format,
								   Uint8 channels, int freq);
}
#endif

namespace {
	void callback_(void* orchestra, Uint8* stream, int len) {
		static_cast<Orchestra*>(orchestra)->callback(stream, len);
	}

#ifdef USE_SDL_SOUND
	// from playsound.c
	typedef struct
	{
		Uint8 *decoded_ptr;
		Uint32 decoded_bytes;
		int predecode;
		int looping;
		int wants_volume_change;
		float volume;
		Uint32 total_seeks;
		Uint32 *seek_list;
		Uint32 seek_index;
		Sint32 bytes_before_next_seek;
	} playsound_global_state;

	volatile playsound_global_state global_state;


	volatile int done_flag = 0;

	Uint32 cvtMsToBytePos(Sound_AudioInfo *info, Uint32 ms)
	{
		/* "frames" == "sample frames" */
		float frames_per_ms = ((float) info->rate) / 1000.0;
		Uint32 frame_offset = (Uint32) (frames_per_ms * ((float) ms));
		Uint32 frame_size = (Uint32) ((info->format & 0xFF) / 8) * info->channels;
		return(frame_offset * frame_size);
	} /* cvtMsToBytePos */

	void do_seek(Sound_Sample *sample)
	{
		Uint32 *seek_list = global_state.seek_list;
		Uint32 seek_index = global_state.seek_index;
		Uint32 total_seeks = global_state.total_seeks;

		fprintf(stdout, "Seeking to %.2d:%.2d:%.4d...\n",
				(int) ((seek_list[seek_index] / 1000) / 60),
				(int) ((seek_list[seek_index] / 1000) % 60),
				(int) ((seek_list[seek_index] % 1000)));

		if (global_state.predecode)
		{
			Uint32 pos = cvtMsToBytePos(&sample->desired, seek_list[seek_index]);
			if (pos > sample->buffer_size)
			{
				fprintf(stderr, "Seek past end of predecoded buffer.\n");
				done_flag = 1;
			} /* if */
			else
			{
				global_state.decoded_ptr = (((Uint8 *) sample->buffer) + pos);
				global_state.decoded_bytes = sample->buffer_size - pos;
			} /* else */
		} /* if */
		else
		{
			if (!Sound_Seek(sample, seek_list[seek_index]))
			{
				fprintf(stderr, "Sound_Seek() failed: %s\n", Sound_GetError());
				done_flag = 1;
			} /* if */
		} /* else */

		seek_index++;
		if (seek_index >= total_seeks)
			global_state.bytes_before_next_seek = -1;  /* no more seeks. */
		else
		{
			global_state.bytes_before_next_seek = cvtMsToBytePos(&sample->desired,
																 seek_list[seek_index]);
			seek_index++;
		} /* else */

		global_state.seek_index = seek_index;
	} /* do_seek */



    /*
	 * This updates (decoded_bytes) and (decoded_ptr) with more audio data,
	 *  taking into account potential looping, seeking and predecoding.
	 */
	int read_more_data(Sound_Sample *sample)
	{
		if (done_flag)              /* probably a sigint; stop trying to read. */
		{
			global_state.decoded_bytes = 0;
			return(0);
		} /* if */

		if ((global_state.bytes_before_next_seek >= 0) &&
			(global_state.decoded_bytes > global_state.bytes_before_next_seek))
		{
			global_state.decoded_bytes = global_state.bytes_before_next_seek;
		} /* if */

		if (global_state.decoded_bytes > 0) /* don't need more data; just return. */
			return(global_state.decoded_bytes);

        /* Need more audio data. See if we're supposed to seek... */
		if ((global_state.bytes_before_next_seek == 0) &&
			(global_state.seek_index < global_state.total_seeks))
		{
			do_seek(sample);  /* do it, baby! */
			return(read_more_data(sample));  /* handle loops conditions. */
		} /* if */

        /* See if there's more to be read... */
		if ( (global_state.bytes_before_next_seek != 0) &&
			 (!(sample->flags & (SOUND_SAMPLEFLAG_ERROR | SOUND_SAMPLEFLAG_EOF))) )
		{
			global_state.decoded_bytes = Sound_Decode(sample);
			if (sample->flags & SOUND_SAMPLEFLAG_ERROR)
			{
				fprintf(stderr, "Error in decoding sound file!\n"
						"  reason: [%s].\n", Sound_GetError());
			} /* if */

			global_state.decoded_ptr = (Uint8* volatile)sample->buffer;
			return(read_more_data(sample));  /* handle loops conditions. */
		} /* if */

		/* No more to be read from stream, but we may want to loop the sample. */

		if (!global_state.looping)
			return(0);

		global_state.looping--;

		global_state.seek_index = 0;
		global_state.bytes_before_next_seek =
			(global_state.total_seeks > 0) ? 0 : -1;

		/* we just need to point predecoded samples to the start of the buffer. */
		if (global_state.predecode)
		{
			global_state.decoded_bytes = sample->buffer_size;
			global_state.decoded_ptr = (Uint8* volatile)sample->buffer;
		} /* if */
		else
		{
			Sound_Rewind(sample);  /* error is checked in recursion. */
		} /* else */

		return(read_more_data(sample));
	} /* read_more_data */

	void memcpy_with_volume(Sound_Sample *sample,
							Uint8 *dst, Uint8 *src, int len)
	{
		int i;
		Uint16 *u16src = NULL;
		Uint16 *u16dst = NULL;
		Sint16 *s16src = NULL;
		Sint16 *s16dst = NULL;
		float volume = global_state.volume;

		if (!global_state.wants_volume_change)
		{
			memcpy(dst, src, len);
			return;
		} /* if */

		/* !!! FIXME: This would be more efficient with a lookup table. */
		switch (sample->desired.format)
		{
        case AUDIO_U8:
            for (i = 0; i < len; i++, src++, dst++)
                *dst = (Uint8) (((float) (*src)) * volume);
            break;

        case AUDIO_S8:
            for (i = 0; i < len; i++, src++, dst++)
                *dst = (Sint8) (((float) (*src)) * volume);
            break;

        case AUDIO_U16LSB:
            u16src = (Uint16 *) src;
            u16dst = (Uint16 *) dst;
            for (i = 0; i < len; i += sizeof (Uint16), u16src++, u16dst++)
            {
                *u16dst = (Uint16) (((float) (SDL_SwapLE16(*u16src))) * volume);
                *u16dst = SDL_SwapLE16(*u16dst);
            } /* for */
            break;

        case AUDIO_S16LSB:
            s16src = (Sint16 *) src;
            s16dst = (Sint16 *) dst;
            for (i = 0; i < len; i += sizeof (Sint16), s16src++, s16dst++)
            {
                *s16dst = (Sint16) (((float) (SDL_SwapLE16(*s16src))) * volume);
                *s16dst = SDL_SwapLE16(*s16dst);
            } /* for */
            break;

        case AUDIO_U16MSB:
            u16src = (Uint16 *) src;
            u16dst = (Uint16 *) dst;
            for (i = 0; i < len; i += sizeof (Uint16), u16src++, u16dst++)
            {
                *u16dst = (Uint16) (((float) (SDL_SwapBE16(*u16src))) * volume);
                *u16dst = SDL_SwapBE16(*u16dst);
            } /* for */
            break;

        case AUDIO_S16MSB:
            s16src = (Sint16 *) src;
            s16dst = (Sint16 *) dst;
            for (i = 0; i < len; i += sizeof (Sint16), s16src++, s16dst++)
            {
                *s16dst = (Sint16) (((float) (SDL_SwapBE16(*s16src))) * volume);
                *s16dst = SDL_SwapBE16(*s16dst);
            } /* for */
            break;
		} /* switch */
	} /* memcpy_with_volume */
#endif
}

void rdft(int, int, double *, int *, double *);

#define DEFAULT_DECODEBUF 16384
#define DEFAULT_AUDIOBUF  4096

void Orchestra::initSound(const std::string& musicFile) {
	// from playsound.c
#ifdef USE_SDL_SOUND
	done_flag = 0;

    Uint32 audio_buffersize;
    Uint32 decode_buffersize;

	if (global_state.seek_list != NULL)
		free((void *) global_state.seek_list);

	memset((void *) &global_state, '\0', sizeof (global_state));
	global_state.volume = 1.0;
	global_state.bytes_before_next_seek = -1;
	audio_buffersize = DEFAULT_AUDIOBUF;
	decode_buffersize = DEFAULT_DECODEBUF;

    Sound_AudioInfo sound_desired;
	sound_desired.format = AUDIO_S16;
	sound_desired.channels = 2;
	sound_desired.rate = 22050;

	sample = Sound_NewSampleFromFile(
		musicFile.c_str(), &sound_desired, decode_buffersize);

	if (sample == 0) {
		error("couldn't load " + musicFile + ": " + Sound_GetError());
	}

    SDL_AudioSpec sdl_desired;

	sdl_desired.freq = 22050;
	sdl_desired.format = AUDIO_S16;
	sdl_desired.channels = 2;
	sdl_desired.samples = 2048;
	sdl_desired.callback = callback_;
	sdl_desired.userdata = this;

	if (SDL_OpenAudio(&sdl_desired, NULL) < 0) {
		error(std::string("cannot open audio device: ") + SDL_GetError());
	}
#else
	Mix_OpenAudio(22050, AUDIO_S16, 2, 2048);

	music_ = Mix_LoadMUS(musicFile.c_str());
	check(music_ != 0,
		  std::string("cannot open ") + musicFile + ": " +
		  SDL_GetError() + "\n");

	Mix_SetPostMix(callback_, this);
#endif
}

Orchestra::Orchestra(const std::string& musicFile, Game* game)
	: game_(game), prev_(0),
/*	  vol_sense_inc_(1.08), vol_sense_dec_(0.92), vol_sense_inc2_(1.3),
	  vol_sense_(1000000 >> 4), vol_sense_min_(100000 >> 4),
      freqAvg_(0), */
	  notBulletTurn_(0),
	  times_(0), pointTimes_(0), maxSum_(0),
	  rankRatio_(0.5),
	  isEasy_(BPO::rc().get<std::string>("difficulty") == "Easy")
{
	std::cout << "orchesra init: " << musicFile << std::endl;

	spectru_.resize(FREQ_NUM);
	spec_.resize(FREQ_NUM);
	smooth_.resize(FREQ_NUM);

	initSound(musicFile);

	seed_ = 1;
	for (size_t i = 0; i < musicFile.length(); i++) {
		seed_ *= musicFile[i];
	}
	std::ifstream is(musicFile.c_str());
	for (int i = 1; i <= 10; i++) {
		is.seekg(1000*i);
		int c = is.get();
		if (c != -1) seed_ *= (c+1);
	}
	init_genrand(seed_);

	std::cout << "orchesra init done" << std::endl;
}

Orchestra::~Orchestra() {
	end();
#ifdef USE_SDL_SOUND
	SDL_CloseAudio();
	Sound_FreeSample(sample);
#else
	Mix_FreeMusic(music_);
	Mix_SetPostMix(0, 0);
	Mix_CloseAudio();
#endif
}

void Orchestra::resetBullets() {
	normalMainBullet_ = BPO::getNormalBullet();
	do {
		normalSubBullet_ = BPO::getEasyBullet();
	} while (normalSubBullet_ == normalMainBullet_);
	do {
		hardBullet_ = BPO::getHardBullet();
	} while (hardBullet_ == normalMainBullet_);
}

void Orchestra::start() {
#ifdef USE_SDL_SOUND
	SDL_PauseAudio(0);
#else
	if (Mix_PlayMusic(music_, 1) != 0) {
		error(std::string("cannot start music: ") + SDL_GetError());
	}
#endif
	resetBullets();
}

void Orchestra::end() {
#ifdef USE_SDL_SOUND
	SDL_PauseAudio(1);
#else
	Mix_HaltMusic();
#endif
}

bool Orchestra::isEnd() {
#ifdef USE_SDL_SOUND
	return done_flag;
#else
	return !Mix_PlayingMusic();
#endif
}

/**
 * 르ꥺΥ⡣
 * ޤ FFT
 * ʬΰΥǡ FREQ_NUM ȿӤ float ˤޤȤ롣
 * λ˺ͤĴ٤Ƥơ
 * ޤǤκͤʿѤ 512ˤʤ褦 int Ѵ
 * 줫餽줾μȿӤ濴˽ΨͿ
 * ûʿѤȤ뤳ȤˤäƳ꤫ʥǡˤ롣
 * ꤫ˤʤեޥȤĴ٤롣
 * եޥȤ롣
 * ǡ餫ե󥰤Ƥȯͤ᤿ꡣ
 *
 * ʾΥ르ꥺŬ˹ͤΤʤΤǡ
 * 򤬤櫓ǤϤޤ
 * äɤˡʤɤФĺй
 */
void Orchestra::callback(Uint8* stream, int len) {
#ifdef USE_SDL_SOUND
    int bw = 0; /* bytes written to stream this time through the callback */

    while (bw < len)
    {
        int cpysize;  /* bytes to copy on this iteration of the loop. */

        if (!read_more_data(sample)) /* read more data, if needed. */
        {
            /* ...there isn't any more data to read! */
            memset(stream + bw, '\0', len - bw);
            done_flag = 1;
            return;
        } /* if */

        /* decoded_bytes and decoder_ptr are updated as necessary... */

        cpysize = len - bw;
        if (cpysize > global_state.decoded_bytes)
            cpysize = global_state.decoded_bytes;

        if (cpysize > 0)
        {
            memcpy_with_volume(sample, stream + bw,
                               (Uint8 *) global_state.decoded_ptr,
                               cpysize);

            bw += cpysize;
            global_state.decoded_ptr += cpysize;
            global_state.decoded_bytes -= cpysize;
            if (global_state.bytes_before_next_seek >= 0)
                global_state.bytes_before_next_seek -= cpysize;
        } /* if */
    } /* while */
#endif

	times_++;
	if (a.size() < (size_t)len+1) {
		a.resize(len+1);
		w.resize(len*5/4);
		ip.resize((int)sqrt(len));
	}

	int i;

	for (i = 0; i < len; i++) {
		a[i] = (Sint8)*stream;
		stream++;
	}

	ip[0] = 0;
	a[0] *= 0.5;
	a[len] = 0;

	// FFT
	rdft(len, 1, &a[0], &ip[0], &w[0]);

	// Ȥϰ̵̣ߤ
	int l = (len >> 4) / FREQ_NUM;
	double* ap = &a[0];
	float* pp = p;
	float max = 0;

	for (i = 0; i < FREQ_NUM; i++) {
		float tmp = 0;
		for (int j = 0; j < l; j++) {
			tmp += fabs(*ap);
			++ap;
		}
		*pp = tmp;
		if (tmp > max) {
			max = tmp;
		}
		++pp;
	}

	maxSum_ += max;
	float maxAvg = maxSum_ / times_;
	pp = p;

	for (i = 0; i < FREQ_NUM_2; i++) {
		int pow = (int)(*pp * SPECTRU_SIZE / maxAvg);
		spec_[i] = pow;
		pp++;
	}

	int formant1 = -1;
	int formant2 = -1;
	int formant1val = 0;
	for (i = PEEK_INT; i < FREQ_NUM_2-PEEK_INT; i++) {
		int p = 0;
		for (int j = -PEEK_INT; j <= PEEK_INT; j++) {
			if (j < 0) {
				p += spec_[i+j] * (PEEK_INT + 1 + j);
			}
			else {
				p += spec_[i+j] * (PEEK_INT + 1 - j);
			}
		}
		p >>= PEEK_INT+1;
		smooth_[i] = p;
		if (formant1val < p) {
			formant1 = i;
			formant1val = p;
		}
	}

	int formant2val = 0;
	int valley = formant1val >> 1;
	int prev = formant1val;
	bool nowFormant1 = true;
	for (i = formant1-1; i >= PEEK_INT; i--) {
		if (prev < smooth_[i] && smooth_[i] < valley) nowFormant1 = false;
		prev = smooth_[i];
		if (!nowFormant1 && formant2val < smooth_[i]) {
			formant2 = i;
			formant2val = smooth_[i];
		}
	}
	prev = formant1val;
	nowFormant1 = true;
	for (i = formant1+1; i < FREQ_NUM_2-PEEK_INT; i++) {
		if (prev < smooth_[i] && smooth_[i] < valley) nowFormant1 = false;
		prev = smooth_[i];
		if (!nowFormant1 && formant2val < smooth_[i]) {
			formant2 = i;
			formant2val = smooth_[i];
		}
	}

/*
  if (formant1 > formant2) {
  std::swap(formant1, formant2);
  std::swap(formant1val, formant2val);
  }
*/

//		printf("%d %d %d %d\n", formant1, formant2, smooth_[formant1], smooth_[formant2]);

	pointTimes_++;

	int point = static_cast<int>(max); //formant1val;// + formant2val << 1;

	double ratioChanging = (times_ > 60) ? 0.4 : 1.0 - times_ * 0.01;

//		std::cout << rankRatio_ << std::endl;;

	histry_.insert(point);
	int order = 1;
	for (std::multiset<int>::reverse_iterator ite = histry_.rbegin();
		 ite != histry_.rend(); ++ite)
	{
		if (*ite < point) break;

		if (order > pointTimes_ / 8 * rankRatio_) {
			notBulletTurn_++;

			if (notBulletTurn_ > 30) {
				std::cout << "change bullet" << std::endl;
				resetBullets();
				//BPO::resetColors();
				notBulletTurn_ = 0;
				if (rankRatio_ < 1) {
					rankRatio_ *= 2 - rankRatio_;
				}
			}

			double baseRatio = (times_ > 50) ? 0.05 : times_ * 0.001;
			if (isEasy_) baseRatio *= 0.5;
			rankRatio_ *= 1 + baseRatio * ratioChanging;

			return;
		}

		order++;
	}

	if (order < pointTimes_ / 16 * rankRatio_) {
		if (formant1 > 30 && formant1val > (formant2val << 1)) {
			std::cout << "hard" << std::endl;
			game_->reserveShot(hardBullet_, game_->getTurn()+1);
			rankRatio_ *= 1 - 0.6 * ratioChanging;
		}
		else {
			std::cout << "main" << std::endl;
			game_->reserveShot(normalMainBullet_, game_->getTurn()+1);
			rankRatio_ *= 1 - 0.35 * ratioChanging;
		}
	}
	else {
		std::cout << "sub" << std::endl;
		game_->reserveShot(normalSubBullet_, game_->getTurn()+1);
		rankRatio_ *= 1 - 0.15 * ratioChanging;
	}
}

